/* -*- mode: c++; tab-width: 8; c-file-style: "gnu" -*- */
/*
  SDL - Simple DirectMedia Layer
  Copyright (C) 1997, 1998, 1999, 2000, 2001, 2002  Sam Lantinga

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Sam Lantinga
  slouken@libsdl.org
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysvideo.cc,v 1.3 2002/09/01 00:37:24 slouken Exp $";
#endif

/* Qtopia based framebuffer implementation */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fb.h>
#include <sys/mman.h>
#include <asm/page.h>	

#include <qapplication.h>
#include <qpe/qpeapplication.h>
#include <qpe/qcopenvelope_qws.h>
#include <qgfx_qws.h>
#include <qwindowsystem_qws.h>
#include <qwidget.h>
#include <qwidgetlist.h>
#include <qdirectpainter_qws.h>

#include "SDL.h"
#include "SDL_timer.h"

#include "SDL_QWin.h"
#include "SDL_sysvideo.h"

extern "C" {

#include "SDL_sysvideo.h"
#include "SDL_sysmouse_c.h"
#include "SDL_sysevents_c.h"
#include "SDL_events_c.h"
#include "SDL_syswm_c.h"
#include "SDL_lowvideo.h"

  // #define QTOPIA_DEBUG
#define QTOPIA_LOG
#define QT_HIDDEN_SIZE	32	/* starting hidden window size */

#ifdef QTOPIA_LOG
  static FILE *st_logfp = NULL;
#endif
  static inline void LOG(char *fmt, ...)
  {
#ifdef QTOPIA_LOG
    va_list ap;

    va_start(ap, fmt);
    vfprintf(st_logfp, fmt, ap);
    if (st_logfp != stderr) {
      fflush(st_logfp);
      vfprintf(stderr, fmt, ap);
    }
    va_end(ap);
#endif
  }

  typedef struct machine_spec {
    const char *manif;
    const char *name;
    int qpe_server_rotation;
    int init_screen_rot;
    int init_key_rot;
  } machine_spec_t;

  typedef enum {
    MACHINE_SHARP_SL5000D,
    MACHINE_SHARP_SL5500,
    MACHINE_SHARP_SLA300,
    MACHINE_SHARP_SLB500,
    MACHINE_SHARP_SLC700,
    MACHINE_SHARP_SLC750,
    MACHINE_SHARP_SLC760,
    MACHINE_SHARP_SLC860,
    MACHINE_SHARP_SL6000,
    MACHINE_SHARP_SLC3000,
    MACHINE_MAX
  } machine_t;

  static const machine_spec_t st_machine_spec[] = {
    { "Sharp", "SL-5000D", 0 },
    { "Sharp", "SL-5500", 0 },
    { "Sharp", "SL-A300", 0 },
    { "Sharp", "SL-B500", 0 },
    { "Sharp", "SL-C700", 0 },
    { "Sharp", "SL-C750", 0 },
    { "Sharp", "SL-C760", 0 },
    { "Sharp", "SL-C860", 0 },
    { "Sharp", "SL-6000", 0 },
    { "Sharp", "SL-C3000", 0 },
  };

/* Name of the environment variable used to invert the screen rotation or not:
   Possible values:
   !=0 : Screen is 270- rotated
   0: Screen is 90- rotated*/
#define SDL_QT_ROTATION_ENV_NAME "SDL_QT_ROTATION"
#define SDL_QT_INVERT_ROTATION_ENV_NAME "SDL_QT_INVERT_ROTATION"

  /* Initialization/Query functions */
  static int QT_VideoInit(_THIS, SDL_PixelFormat *vformat);
  static SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
  static SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current, int width, int height, int bpp, Uint32 flags);
  static void QT_UpdateMouse(_THIS);
  static int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
  static void QT_VideoQuit(_THIS);

  /* Hardware surface functions */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface);
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface);
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface);

  static int QT_ToggleFullScreen(_THIS, int fullscreen);

  static int QT_IconifyWindow(_THIS);
  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode);

  static int console_fd;
  struct fb_var_screeninfo saved_vinfo;
  int fb_hwrot;
  int fb_direct;
  static int isKHloaded;
  bool is_VGA_machine;
 
#define W100FB_CONFIG          0x57415200 /* WAL\00 */
#define W100INIT_ITEM          0
#define W100INIT_ALL           1
#define W100INIT_ITEM_WITH_VAL 2
#define W100FB_CONFIG_EX       0x57415202 /* WAL\02 */


  /* FB driver bootstrap functions */

  static int QT_Available(void)
  {
    return(1);
  }

  static void QT_DeleteDevice(SDL_VideoDevice *device)
  {
    free(device->hidden);
    free(device);
  }

  static SDL_VideoDevice *QT_CreateDevice(int devindex)
  {
    SDL_VideoDevice *device;

    /* Initialize all variables that we clean on shutdown */
    device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
    if ( device ) {
      memset(device, 0, (sizeof *device));
      device->hidden = (struct SDL_PrivateVideoData *)
	malloc((sizeof *device->hidden));
    }
    if ( (device == NULL) || (device->hidden == NULL) ) {
      SDL_OutOfMemory();
      if ( device ) {
	free(device);
      }
      return(0);
    }
    memset(device->hidden, 0, (sizeof *device->hidden));

    /* Set the function pointers */
    device->VideoInit = QT_VideoInit;
    device->ListModes = QT_ListModes;
    device->SetVideoMode = QT_SetVideoMode;
    device->UpdateMouse = QT_UpdateMouse;
    device->SetColors = QT_SetColors;
    device->UpdateRects = NULL;
    device->VideoQuit = QT_VideoQuit;
    device->AllocHWSurface = QT_AllocHWSurface;
    device->CheckHWBlit = NULL;
    device->FillHWRect = NULL;
    device->SetHWColorKey = NULL;
    device->SetHWAlpha = NULL;
    device->LockHWSurface = QT_LockHWSurface;
    device->UnlockHWSurface = QT_UnlockHWSurface;
    device->FlipHWSurface = NULL;
    device->FreeHWSurface = QT_FreeHWSurface;
    device->SetIcon = NULL;
    device->SetCaption = QT_SetWMCaption;
    device->IconifyWindow = QT_IconifyWindow;
    device->GrabInput = QT_GrabInput;
    device->GetWMInfo = NULL;
    device->FreeWMCursor = QT_FreeWMCursor;
    device->CreateWMCursor = QT_CreateWMCursor;
    device->ShowWMCursor = QT_ShowWMCursor;
    device->WarpWMCursor = QT_WarpWMCursor;
    device->InitOSKeymap = QT_InitOSKeymap;
    device->PumpEvents = QT_PumpEvents;

    device->free = QT_DeleteDevice;
    device->ToggleFullScreen = QT_ToggleFullScreen;

    /* Set the driver flags */
    device->handles_any_size = 0;
	
    return device;
  }

  VideoBootStrap Qtopia_bootstrap = {
    "qtopia", "Qtopia / QPE graphics",
    QT_Available, QT_CreateDevice
  };

  /* Function to sort the display_list */
  static int CompareModes(const void *A, const void *B)
  {
#if 0
    const display_mode *a = (display_mode *)A;
    const display_mode *b = (display_mode *)B;

    if ( a->space == b->space ) {
      return((b->virtual_width*b->virtual_height)-
	     (a->virtual_width*a->virtual_height));
    } else {
      return(ColorSpaceToBitsPerPixel(b->space)-
	     ColorSpaceToBitsPerPixel(a->space));
    }
#endif
    return 0;
  }

  /* Yes, this isn't the fastest it could be, but it works nicely */
  static int QT_AddMode(_THIS, int index, unsigned int w, unsigned int h)
  {
    SDL_Rect *mode;
    int i;
    int next_mode;

    /* Check to see if we already have this mode */
    if ( SDL_nummodes[index] > 0 ) {
      for ( i=SDL_nummodes[index]-1; i >= 0; --i ) {
	mode = SDL_modelist[index][i];
	if ( (mode->w == w) && (mode->h == h) ) {
	  return(0);
	}
      }
    }

    /* Set up the new video mode rectangle */
    mode = (SDL_Rect *)malloc(sizeof *mode);
    if ( mode == NULL ) {
      SDL_OutOfMemory();
      return(-1);
    }
    mode->x = 0;
    mode->y = 0;
    mode->w = w;
    mode->h = h;
#ifdef QTOPIA_DEBUG
    fprintf(stderr, "Adding mode %dx%d at %d bytes per pixel\n", w, h, index+1);
#endif

    /* Allocate the new list of modes, and fill in the new mode */
    next_mode = SDL_nummodes[index];
    SDL_modelist[index] = (SDL_Rect **)
      realloc(SDL_modelist[index], (1+next_mode+1)*sizeof(SDL_Rect *));
    if ( SDL_modelist[index] == NULL ) {
      SDL_OutOfMemory();
      SDL_nummodes[index] = 0;
      free(mode);
      return(-1);
    }
    SDL_modelist[index][next_mode] = mode;
    SDL_modelist[index][next_mode+1] = NULL;
    SDL_nummodes[index]++;

    return(0);
  }

  int QT_VideoInit(_THIS, SDL_PixelFormat *vformat)
  {
    const char *SDL_fbdev;
    struct fb_var_screeninfo vinfo;

     /* Initialize the QPE Application  */
     /* Determine the screen depth */
    vformat->BitsPerPixel = QPixmap::defaultDepth();

    // For now we hardcode the current depth because anything else
    // might as well be emulated by SDL rather than by Qtopia.

    //frame buffer device open.
    SDL_fbdev = getenv("SDL_FBDEV");
    if ( SDL_fbdev == NULL ) {
    	SDL_fbdev = "/dev/fb0";
    }
    console_fd = open(SDL_fbdev, O_RDWR, 0);
    if ( console_fd < 0 ) {
    	SDL_SetError("Unable to open %s", SDL_fbdev);
    	return(-1);
    }

    if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
    	SDL_SetError("Couldn't get console pixel format");
    	QT_VideoQuit(_this);
    	return(-1);
    }
    saved_vinfo = vinfo;

    QSize desktop_size = qApp->desktop()->size();
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.width(), desktop_size.height());
    QT_AddMode(_this, ((vformat->BitsPerPixel+7)/8)-1,
	       desktop_size.height(), desktop_size.width());

    /* Create the window / widget */
    SDL_Win = new SDL_QWin(QSize(QT_HIDDEN_SIZE, QT_HIDDEN_SIZE));
    ((QPEApplication*)qApp)->showMainWidget(SDL_Win);
    /* Fill in some window manager capabilities */
    _this->info.wm_available = 0;

#ifdef QTOPIA_LOG
    st_logfp = fopen("/tmp/sdl-qt-debug", "w");
    if (!st_logfp)
      st_logfp = stderr;
#endif

    QT_GrabInput(_this, SDL_GRAB_ON);

    // keyhelper loading for SDL

    isKHloaded=(fopen("/home/zaurus/Settings/keyhelper_SDL.xml","r") != NULL) ? 1 : 0;
    if (isKHloaded) {
	system("qcop QPE/KeyHelper \"reload(QString)\" keyhelper_SDL.xml"); //QcopEnvelope can't use. 
	LOG("keyhelper_SDL.xml loaded\n");
    }
    return(0);
  }

  /* We support any dimension at our bit-depth */
  SDL_Rect **QT_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
  {
    SDL_Rect **modes;

    modes = ((SDL_Rect **)0);
    if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
      modes = SDL_modelist[((format->BitsPerPixel+7)/8)-1];
    } else {
      if ( format->BitsPerPixel ==
	   _this->screen->format->BitsPerPixel ) {
	modes = ((SDL_Rect **)-1);
      }
    }
    return(modes);
  }

  /* Various screen update functions available */
  static void QT_NormalUpdate(_THIS, int numrects, SDL_Rect *rects);
  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects);


  static int QT_SetFullScreen(_THIS, SDL_Surface *screen, int fullscreen)
  {
    return -1;
  }

  static int QT_ToggleFullScreen(_THIS, int fullscreen)
  {
    return -1;
  }

  static machine_t QT_GetMachine(_THIS)
  {
    FILE *fp;
    machine_t machine = MACHINE_SHARP_SL5000D;
    char buf[512];

    fp = fopen("/proc/deviceinfo/product", "rb");
    if (fp) {
      if (fread(buf, 1, sizeof(buf), fp) > 0) {
//        LOG("QT_GetMachine: /proc/deviceinfo/product is %s\n", buf);
        if (strncmp(buf, "SL-A300", 7) == 0)
          machine = MACHINE_SHARP_SLA300;
        else if (strncmp(buf, "SL-5500", 7) == 0)
          machine = MACHINE_SHARP_SL5500;
        else if (strncmp(buf, "SL-B500", 7) == 0)
          machine = MACHINE_SHARP_SLB500;
        else if (strncmp(buf, "SL-C700", 7) == 0)
          machine = MACHINE_SHARP_SLC700;
        else if (strncmp(buf, "SL-C750", 7) == 0)
          machine = MACHINE_SHARP_SLC750;
        else if (strncmp(buf, "SL-C760", 7) == 0)
          machine = MACHINE_SHARP_SLC760;
        else if (strncmp(buf, "SL-C860", 7) == 0)
          machine = MACHINE_SHARP_SLC860;
        else if (strncmp(buf, "SL-6000", 7) == 0)
          machine = MACHINE_SHARP_SL6000;
        else if (strncmp(buf, "SL-C3000", 8) == 0)
          machine = MACHINE_SHARP_SLC3000;
      }
      fclose(fp);
    } else {
      LOG("QT_GetMachine: Couldn't read /proc/deviceinfo/product.\n");
      LOG(" Now set machine variable to default (SL-5000D)\n");
    }

    LOG(" detected machine is '%s %s'\n",
        st_machine_spec[machine].manif, st_machine_spec[machine].name);
    return machine;
  }

  static void QT_GetQteServerSpec(_THIS, int *rotation, bool *isQvga)
  {
    const char *user;
    char buf[FILENAME_MAX];
    FILE *fp;
    int rot = 0, is_qvga = 0;

    user = getenv("USER");
    snprintf(buf, sizeof(buf), "/tmp/qtembedded-%s/QtEmbedded-0.spec",
             user ? user : "root");
    LOG("QT_GetRotation: Read spec from '%s'\n", buf);

    fp = fopen(buf, "rb");
    if (fp) {
      int size;
      if ((size = fread(buf, 1, sizeof(buf) - 1, fp)) > 0) {
        buf[size] = '\0';
        LOG(" spec is '%s'\n", buf);
        // get rotation value
        if (strstr(buf, "Rot")) {
          rot = atoi(strstr(buf, "Rot") + 3);
          rot /= 90;
          if (rot < 0 || rot > 3) {
            rot = 0;
          }
        } else
          rot = 0;

        // get qvga mode in SL-C700
        if (strstr(buf, "Qvga"))
          is_qvga = 1;
      }
      fclose(fp);
    }

    LOG(" Rot=%d, Qvga=%d\n", rot, is_qvga);

    if (rotation)
      *rotation = rot;
    if (isQvga)
      *isQvga = is_qvga;
  }

  /* FIXME: check return values and cleanup here */
  SDL_Surface *QT_SetVideoMode(_THIS, SDL_Surface *current,
			       int width, int height, int bpp, Uint32 flags)
  {
    QImage *qimage;
    QSize qteSize = qApp->desktop()->size();
    QSize fbSize;
    QSize userSize;
    int fb_xres;
    int fb_yres;
    struct fb_var_screeninfo vinfo;
    struct fb_fix_screeninfo finfo;
    int tmp_ioctl_data;
    char fb_size_fix = ' ';
    bool isW100patch_kernel = false;
    int numFb_Size;

    int mapped_memlen;
    int mapped_offset;
    void *mapped_mem;
    int fb_offset;


    machine_t machine = QT_GetMachine(_this);
    machine_spec_t machineSpec = st_machine_spec[machine];
    int qteRotation, userRotation, sdlRotation;
    int qteKeyRotation, sdlKeyRotation;
    bool isQteQvga;

    if (machine == MACHINE_SHARP_SLC700 ||
        machine == MACHINE_SHARP_SLC750 ||
        machine == MACHINE_SHARP_SLC760 ||
        machine == MACHINE_SHARP_SLC860 ||
        machine == MACHINE_SHARP_SL6000 ||
        machine == MACHINE_SHARP_SLC3000 ) {
	is_VGA_machine = true;
    }else {
	is_VGA_machine = false;
    }

    // qte Ǥβž٤
    QT_GetQteServerSpec(_this, &qteRotation, &isQteQvga);

    // SL-C700 Υ
    bool isInputStyle = false;
    if (is_VGA_machine) {
      int status = system("/home/QtPalmtop/bin/chkhinge");
      if (WEXITSTATUS(status) != 2)
        isInputStyle = true;
      LOG("QT_SetVideoMode: SL-C700 Style is %s\n",
          isInputStyle ? "Input style" : "View style");
    }

    
    // specity screen setting
    if (is_VGA_machine) {

      // w100 rotation pached kernel check
      isW100patch_kernel = (fopen("/proc/driver/w100/rotation","r") != NULL) ? true : false;
      LOG("Your Kernel is %s\n",isW100patch_kernel ? "Special Kernel with w100" : "Normal Kernel or without w100");

      const char *envFb_Size = getenv("SDL_FB_SIZE");
      if (envFb_Size !=NULL) {
	numFb_Size=sscanf(envFb_Size,"%dx%d%c",&fb_xres,&fb_yres,&fb_size_fix);
      } else {
	numFb_Size=0;
      }

      if (numFb_Size >=2) {

	// specity screen mode
	if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
		SDL_SetError("Couldn't get console pixel format");
		QT_VideoQuit(_this);
		return(NULL);
	}

	vinfo.xres = fb_xres;
	vinfo.xres_virtual = fb_xres;
	vinfo.yres = fb_yres;
	vinfo.yres_virtual = fb_yres;

	if (fb_size_fix == '@') {
		width = fb_xres;
		height = fb_yres;
	}

	if ( ioctl(console_fd, FBIOPUT_VSCREENINFO, &vinfo) < 0 ) {
		SDL_SetError("Couldn't set console screen info");
		return(NULL);
	}


	if (isW100patch_kernel) {
		qteSize.setWidth(vinfo.xres);
		qteSize.setHeight(vinfo.yres);
		qteRotation=0;
	}else if(vinfo.xres == 320){
		qteSize.setWidth(vinfo.xres);
		qteSize.setHeight(vinfo.yres);
		qteRotation=2;
	}else {
		qteSize.setWidth(vinfo.yres);
		qteSize.setHeight(vinfo.xres);
		qteRotation=3;
	}			

	if (isW100patch_kernel && (vinfo.xres == 320) && (vinfo.yres == 240)) {
		tmp_ioctl_data=121;
		ioctl(console_fd, W100FB_CONFIG_EX, &tmp_ioctl_data);
		fb_hwrot=1;	
	}

	LOG("FBVideoMode: %dx%d%c\n", vinfo.xres, vinfo.yres,fb_size_fix );

      }else {

	// auto screen mode	
	if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
		SDL_SetError("Couldn't get console pixel format");
		QT_VideoQuit(_this);
		return(NULL);
	}

	if (isW100patch_kernel && (isInputStyle == false) && (vinfo.xres == 240) && (vinfo.yres == 320)) {
		vinfo.xres = 240;
		vinfo.xres_virtual = 240;
		vinfo.yres = 320;
		vinfo.yres_virtual = 320;
		qteRotation=0;
	}else if ((width <= 320) && (height <= 240)) {
		vinfo.xres = 320;
		vinfo.xres_virtual = 320;
		vinfo.yres = 240;
		vinfo.yres_virtual = 240;
	}else if (isW100patch_kernel && (isInputStyle == true)){
		vinfo.xres = 640;
		vinfo.xres_virtual = 640;
		vinfo.yres = 480;
		vinfo.yres_virtual = 480;
		qteRotation=0;
	}else {
		vinfo.xres = 480;
		vinfo.xres_virtual = 480;
		vinfo.yres = 640;
		vinfo.yres_virtual = 640;
	}

		if ( ioctl(console_fd, FBIOPUT_VSCREENINFO, &vinfo) < 0 ) {
		SDL_SetError("Couldn't set console screen info");
		return(NULL);
	}

	if (isW100patch_kernel) {
		qteSize.setWidth(vinfo.xres);
		qteSize.setHeight(vinfo.yres);
		qteRotation=0;
		fb_direct=1;
	}else if(vinfo.xres == 320){
		qteSize.setWidth(vinfo.xres);
		qteSize.setHeight(vinfo.yres);
		qteRotation=2;
	}else {
		qteSize.setWidth(vinfo.yres);
		qteSize.setHeight(vinfo.xres);
		qteRotation=3;
	}			

	if (isW100patch_kernel && (vinfo.xres == 320) && (vinfo.yres == 240)) {
		tmp_ioctl_data=121;
		ioctl(console_fd, W100FB_CONFIG_EX, &tmp_ioctl_data);	
		fb_hwrot=1;
		qteRotation=0;
	}

	LOG("FBVideoMode: %dx%d%c\n", vinfo.xres, vinfo.yres,fb_size_fix );
      }
    }

    // direct paint setting
    const char *envFb_Direct = getenv("SDL_FB_DIRECT");
    if (envFb_Direct !=NULL)
	fb_direct = envFb_Direct ? atoi(envFb_Direct) : -1;

    if (fb_direct == 1){
	if ( ioctl(console_fd, FBIOGET_FSCREENINFO, &finfo) < 0 ) {
		SDL_SetError("Couldn't get console hardware info");
		QT_VideoQuit(_this);
		return(NULL);
    	}
        if ( ioctl(console_fd, FBIOGET_VSCREENINFO, &vinfo) < 0 ) {
		SDL_SetError("Couldn't get console pixel format");
		QT_VideoQuit(_this);
                return(NULL);
        }
	LOG("Direct paint mode\n");
    }

    // hack for SL-5500
    if (machine == MACHINE_SHARP_SL5500)
      qteRotation = 3;

    // SL-B500 ѥϥå(Ȥꤢ)
    if (machine == MACHINE_SHARP_SLB500)
      qteRotation = 3;

    // ե졼ХåեΥ
    if (qteRotation & 1) {
      fbSize.setWidth(qteSize.height());
      fbSize.setHeight(qteSize.width());
    } else {
      fbSize = qteSize;
    }

    //
    const char *envRotStr = getenv(SDL_QT_ROTATION_ENV_NAME);
    userRotation = envRotStr ? atoi(envRotStr) : -1;
    if ((userRotation >= 0 ? userRotation : qteRotation) & 1) {
      userSize.setWidth(fbSize.height());
      userSize.setHeight(fbSize.width());
    } else {
      userSize = fbSize;
    }

    // ޤ׵ᤵ줿ɥ桼˾ž٤Υ꡼
    // ޤΤå
    if (width <= userSize.width() && height <= userSize.height()) {
      // Ķѿǲž٤ꤵƤСͥ褹
      if (userRotation >= 0)
        sdlRotation = userRotation;
      else {
        // ǡλ̵꤬ΤǤ qte βž٤˹碌
        // ʤɡSL-C700 ǤϤ礤ʣ
	if (is_VGA_machine && (fbSize.width() == 320) && (fbSize.height() == 240)) {
          if (isInputStyle)
            sdlRotation = 2;
          else
            sdlRotation = qteRotation;
        } else {
          sdlRotation = qteRotation;
        }
      }
    } else if (width <= fbSize.width() && height <= fbSize.height()) {
      sdlRotation = 0;
      if (is_VGA_machine && (fbSize.width() == 320) && (fbSize.height() == 240) && isInputStyle)
        sdlRotation = 2;
    } else if (width <= fbSize.height() && height <= fbSize.width()) {
      sdlRotation = 1;
    } else {
      SDL_SetError("Unsupported resolution, %dx%d\n", width, height);
      return NULL;
    }

    if (fb_hwrot == 1) 
	sdlRotation = 0;

    if (getenv(SDL_QT_INVERT_ROTATION_ENV_NAME) != NULL) {
      sdlRotation = (sdlRotation + 2) & 3;
    }

    LOG("QT_SetVideoMode: argSize=%dx%d\n", width, height);
    LOG("QT_SetVideoMode: qteSize=%dx%d\n",
        qteSize.width(), qteSize.height());
    LOG("QT_SetVideoMode: fbSize=%dx%d\n",
        fbSize.width(), fbSize.height());
    LOG("QT_SetVideoMode: qteRotation=%d\n", qteRotation);
    LOG("QT_SetVideoMode: userRotation=%d\n", userRotation);
    LOG("QT_SetVideoMode: sdlRotation=%d\n", sdlRotation);

    current->flags = 0;//SDL_FULLSCREEN; // We always run fullscreen.
    current->w = width;
    current->h = height;
    SDL_Win->setScreenRotation(sdlRotation, qteRotation);
    SDL_Win->setFBSize(fbSize);

    if (machine == MACHINE_SHARP_SLA300)
      SDL_Win->setSLA300InputFix(true);

    // keyboard rotation
    qteKeyRotation = qteRotation;
//     fprintf(stderr, "%d\n", (machine == MACHINE_SHARP_SLC700));
//     fprintf(stderr, "%d\n", isQteQvga);
    if (isQteQvga && is_VGA_machine)
      qteKeyRotation = 3;
    else if (machine == MACHINE_SHARP_SLB500)
      qteKeyRotation = 3;

    if (isQteQvga && fb_hwrot == 1) 
      qteKeyRotation = 1;

    sdlKeyRotation = sdlRotation;

    SDL_Win->setKeyRotation(sdlKeyRotation, qteKeyRotation);

    LOG("QT_SetVideoMode: qteKeyRotation=%d\n", qteKeyRotation);
    LOG("QT_SetVideoMode: sdlKeyRotation=%d\n", sdlKeyRotation);

    if ( flags & SDL_OPENGL ) {
      SDL_SetError("OpenGL not supported");
      return(NULL);
    }
    /* Create the QImage framebuffer */

    // frame buffer memory mapping    
    if (fb_direct == 1) {
	current->pitch = finfo.line_length;
        current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE);
        mapped_offset = (((long)finfo.smem_start) -
	                (((long)finfo.smem_start)&~(PAGE_SIZE-1)));
	mapped_memlen = finfo.smem_len+mapped_offset;
        if(console_fd >0 ) {
		mapped_mem = mmap(NULL, mapped_memlen,PROT_READ|PROT_WRITE, MAP_SHARED, console_fd, 0);
		fb_offset=(vinfo.xres-width)+(vinfo.yres-height)*vinfo.xres;
		current->pixels = (void *)((int )mapped_mem+fb_offset);
                _this->UpdateRects = QT_DirectUpdate;
		LOG("Using FrameBuffer");
    	} else {
	        QDirectPainter *my_painter = new QDirectPainter(SDL_Win);
    		current->pixels = (void *)my_painter->frameBuffer();
                _this->UpdateRects = QT_DirectUpdate;
		LOG("Using DirectPainter\n");
    	}
    } else {
	qimage = new QImage(current->w, current->h, bpp);
	if (qimage->isNull()) {
		SDL_SetError("Couldn't create screen bitmap");
		delete qimage;
		return(NULL);
	}
	current->pitch = qimage->bytesPerLine();
	current->pixels = (void *)qimage->bits();
        _this->UpdateRects = QT_NormalUpdate;
	SDL_Win->setImage(qimage);
//	SDL_Win->setFullscreen(true);  //comment out to non update taskbar
	LOG("Using QImage\n");
    }


    //     fprintf(stderr,"QT_SetVideoMode() qImage:%dx%d %d\n",
    // 			qimage->width(), qimage->height(), qimage->bytesPerLine());
    /* We're done */
    return(current);
  }

  /* Update the current mouse state and position */
  void QT_UpdateMouse(_THIS)
  {
    QPoint point(-1, -1);
    if ( SDL_Win->isActiveWindow() ) {
      point = SDL_Win->mousePos();
    }
    
    if ( (point.x() >= 0) && (point.x() < SDL_VideoSurface->w) &&
	 (point.y() >= 0) && (point.y() < SDL_VideoSurface->h) ) {
      SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
      SDL_PrivateMouseMotion(0, 0,
			     (Sint16)point.x(), (Sint16)point.y());
    } else {
      SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
    }
  }

  /* We don't actually allow hardware surfaces other than the main one */
  static int QT_AllocHWSurface(_THIS, SDL_Surface *surface)
  {
    return(-1);
  }
  static void QT_FreeHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }
  static int QT_LockHWSurface(_THIS, SDL_Surface *surface)
  {
    return(0);
  }
  static void QT_UnlockHWSurface(_THIS, SDL_Surface *surface)
  {
    return;
  }

  static void QT_NormalUpdate(_THIS, int numrects, SDL_Rect *rects)
  {
    if(SDL_Win->lockScreen()) {
      for(int i=0; i<numrects; ++i ) {
	    QRect rect(rects[i].x, rects[i].y,
                  rects[i].w, rects[i].h);
       SDL_Win->repaintRect(rect);
      }
      SDL_Win->unlockScreen();
    }
  }

  static void QT_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
  {

  }

  /* Is the system palette settable? */
  int QT_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
  {
    return -1;
  }

  void QT_VideoQuit(_THIS)
  {
    // This is dumb, but if I free this, the app doesn't exit correctly.
    // Of course, this will leak memory if init video is done more than once.
    // Sucks but such is life.
    
    //    -- David Hedbor
    //    delete SDL_Win; 
    //    SDL_Win = 0;

    if ( console_fd > 0 ) {
    	/* Restore the original video mode and palette */
	if (fb_hwrot == 1) {
		int tmp_ioctl_data=120;
		ioctl(console_fd, W100FB_CONFIG_EX, &tmp_ioctl_data);	
	}

    	ioctl(console_fd, FBIOPUT_VSCREENINFO, &saved_vinfo);

    	/* We're all done with the framebuffer */
    	close(console_fd);
    	console_fd = -1;
    }

    _this->screen->pixels = NULL;
    QT_GrabInput(_this, SDL_GRAB_OFF);
/*
    {
      QCopEnvelope e("QPE/KeyHelper", "repeater(int)");
      e << 1;
    }
*/
    if (isKHloaded)
	system("qcop QPE/KeyHelper \"reload()\""); //QcopEnvelope can't use. 

  }



  static int QT_IconifyWindow(_THIS) {
    SDL_Win->hide();
    
    return true;
  }

  static SDL_GrabMode QT_GrabInput(_THIS, SDL_GrabMode mode) {
    if(mode == SDL_GRAB_OFF) {
      QPEApplication::grabKeyboard();
      qApp->processEvents();
      QPEApplication::ungrabKeyboard();
    } else {
      QPEApplication::grabKeyboard();
    }
    qApp->processEvents();
    return mode;
  }

}; /* Extern C */
